/*
 * AvrStamp
 * A simple plugable and breadboardable board
 *
 * Copyright (C) 2012  B.Stultiens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdlib.h>
#include <stdint.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>

#include <defs.h>
#include <clock.h>
#include <usart.h>

#define WHEEL_TICKS	1000
#define BLINKY_TICKS	333

static const char wheelchars[4] PROGMEM = {'|', '/', '-', '\\'};

static void delay_ms(uint8_t n)
{
	while(n--) {
		for(uint16_t i = 0; i < 3200; i++)
			asm (" nop ");
	}
}

int main(void)
{
	uint8_t wcnt;

	cli();

	MCUCR = 0;		/* Standby mode, no external ints */

	PORTA = 0;
	PORTB = 0;
	PORTC = 0;
	PORTD = 0;
	DDRA = _BV(0);		/* PA0 output for LED */

	wcnt = 0;		/* TIMER_WHEEL character display */

	clock_init();		/* Setup the clock so we may use timing */
	usart_init();		/* Now we can use serial output */

	sei();			/* Enable interrupts */

	/*
	 * Delay before sending serial data to get the host to settle after serial port
	 * init. There may be a change of TX line level that causes framing errors or
	 * break events. The line must be idle for at least one character period.
	 */
	delay_ms(20);

	/* Flush junk and say we are alive */
	usart_puts(PSTR("\n\nBooting...\n"));

	/* Setup the timers which get things done */
	clock_timer_set(TIMER_BLINKY, BLINKY_TICKS);
	clock_timer_set(TIMER_WHEEL, WHEEL_TICKS);

	usart_puts(PSTR("\nPress any key to show uptime and input.\n"));

	for(;;) {
		if(clock_timer_expired(TIMER_BLINKY)) {
			clock_timer_set(TIMER_BLINKY, BLINKY_TICKS);
			tglbit(PORTA, 0);
		}

		if(clock_timer_expired(TIMER_WHEEL)) {
			clock_timer_set(TIMER_WHEEL, WHEEL_TICKS);
			usart_putc(pgm_read_byte(&wheelchars[wcnt++]));
			usart_putc(0x08);
			wcnt %= sizeof(wheelchars);

		}

		if(usart_hasdata()) {
			uint16_t m = clock_msecs;
			usart_putdec_u32(clock_secs);
			usart_putc('.');
			if(m < 100)
				usart_putc('0');
			if(m < 10)
				usart_putc('0');
			usart_putdec_u16(m);
			usart_puts(PSTR(": You pressed a key: 0x"));
			usart_puthex(usart_getc());
			usart_puts(PSTR("\r\n"));
		}

		/* Goto sleep while we have nothing to do */
		sleep_enable();
		sleep_cpu();
	}
}
