/*
 * AvrStamp
 * A simple plugable and breadboardable board
 *
 * Copyright (C) 2012  B.Stultiens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/atomic.h>

#include "defs.h"
#include "usart.h"

#define USART_NO_DROP	1	/* Define if no serial chars should be dropped */
#define USART_NL_CRNL	1	/* Define if NL should be converted to CRNL */

#define USART_TXSIZE	_BV(6)	/* Must be power of 2 (and <= 128) */
#define USART_RXSIZE	_BV(4)	/* Must be power of 2 (and <= 128) */

/*
 * FIXME: <util/setbaud.h> seems to miss the USE_X2, resulting in
 * the wrong speed. Don't know why (yet), but doing it here manually
 * fixes the immediate problem.
 */
#ifndef BAUD
#define BAUD		115200
#endif
#define UBRR_VALUE	((unsigned long)(F_CPU) / (8UL * (unsigned long)(BAUD)) - 1UL)
#define UBRRH_VALUE	(UBRR_VALUE >> 8)
#define UBRRL_VALUE	(UBRR_VALUE & 0xff)
#define USE_X2		1

static volatile char usart_txdata[USART_TXSIZE];
static volatile uint8_t usart_txstart;
static volatile uint8_t usart_txend;
static volatile uint8_t usart_txcnt;
static volatile char usart_rxdata[USART_RXSIZE];
static volatile uint8_t usart_rxstart;
static volatile uint8_t usart_rxend;
static volatile uint8_t usart_rxcnt;

void usart_init(void)
{
	/* Setup USART */
	UBRRH = UBRRH_VALUE & ~_BV(URSEL);
	UBRRL = UBRRL_VALUE;
#if USE_X2
	UCSRA = _BV(U2X);	/* Double speed */
#else
	UCSRA = 0;		/* No double speed */
#endif
	UCSRB = _BV(RXCIE) | _BV(RXEN) | _BV(TXEN) | _BV(UDRIE);
	UCSRC = _BV(URSEL) | _BV(UCSZ0) | _BV(UCSZ1);		/* Async, no parity, 1 stopbit, 8 databits */

	usart_drain();			/* Clear usart data buffers */
	usart_flush();
	/* FIXME: setup stdio so we can use printf() */
}

void usart_putc(char c)
{
#ifdef USART_NL_CRNL
	if(c == '\n')
		usart_putc('\r');	/* Insert CR recursively */
#endif
#ifdef USART_NO_DROP
	/* Stall until space available */
	while(usart_txcnt >= USART_TXSIZE)
		;
#else
	if(usart_txcnt >= USART_TXSIZE)
		return;
#endif
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
		if(!usart_txcnt && bit_is_set(UCSRA, UDRE)) {
			/* Nothing in the buffer and space in the register */
			UDR = c;
		} else {
			usart_txdata[usart_txend++] = c;
			usart_txend &= (USART_TXSIZE-1);
			usart_txcnt++;
			setbit(UCSRB, UDRIE);	/* Enable interrupt if this is the first buffered byte */
		}
	}
}

void usart_puts(const char * PROGMEM s)
{
	char c;
	while((c = pgm_read_byte(s++)))
		usart_putc(c);
}

/*
 * Flush the USART transmitter
 */
void usart_flush(void)
{
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
		usart_txstart = usart_txend = usart_txcnt = 0;
		clrbit(UCSRB, UDRIE);
	}
}

/*
 * Drain the USART receiver
 */
void usart_drain(void)
{
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
		usart_rxstart = usart_rxend = usart_rxcnt = 0;
	}
}

uint8_t usart_hasdata(void)
{
	return usart_rxcnt;
}

char usart_getc(void)
{
	char c;
	while(!usart_rxcnt)
		;
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
		c = usart_rxdata[usart_rxstart++];
		usart_rxstart &= (USART_RXSIZE-1);
		usart_rxcnt--;
	}
	return c;
}

void usart_inject(char c)
{
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
		if(usart_rxcnt < USART_RXSIZE) {
			/* There is room in the buffer */
			usart_rxdata[usart_rxend++] = c;
			usart_rxend &= (USART_RXSIZE-1);
			usart_rxcnt++;
		}
	}
}

#define put_nibble(n)	usart_putc((n) <= 9 ? (n)+'0' : (n)+ '0'+7)
void usart_puthex(uint8_t h)
{
	uint8_t c = h & 0x0f;
	h >>= 4;
	put_nibble(h);
	put_nibble(c);
}

void usart_puthex_u16(uint16_t h)
{
	usart_puthex(h >> 8);
	usart_puthex(h);
}

void usart_puthex_u32(uint32_t h)
{
	usart_puthex_u16(h >> 16);
	usart_puthex_u16(h);
}


void usart_putdec(uint8_t d)
{
	if(d >= 100) {
		usart_putc(d / 100 + '0');
		d %= 100;
		if(d < 10)
			usart_putc('0');
	}
	if(d >= 10) {
		usart_putc(d / 10 + '0');
		d %= 10;
	}
	usart_putc(d + '0');
}

void usart_putdec_u16(uint16_t v)
{
	char buf[6];
	uint8_t c = 0;
	do {
		buf[c++] = (v % 10) + '0';
		v /= 10;
	} while (v);
	do {
		usart_putc(buf[--c]);
	} while(c);
}

void usart_putdec_u32(uint32_t v)
{
	char buf[11];
	uint8_t c = 0;
	do {
		buf[c++] = (v % 10) + '0';
		v /= 10;
	} while (v);
	do {
		usart_putc(buf[--c]);
	} while(c);
}
/*
 * USART transmit data reg empty interrupt
 */
ISR(USART_UDRE_vect)
{
	if(usart_txcnt) {
		UDR = usart_txdata[usart_txstart++];
		usart_txstart &= (USART_TXSIZE-1);
		usart_txcnt--;
	}
	if(!usart_txcnt)
		clrbit(UCSRB, UDRIE);	/* No more data, kill interrupt */
}
/*
 * USART Receive interrupt
 */
ISR(USART_RXC_vect)
{
	uint8_t err = UCSRA;	/* Read error flags */
	char ch = UDR;		/* Read the data */
	if(!(err & (_BV(FE) | _BV(DOR) | _BV(PE))) && usart_rxcnt < USART_RXSIZE) {
		/* No error and there is room in the buffer */
		usart_rxdata[usart_rxend++] = ch;
		usart_rxend &= (USART_RXSIZE-1);
		usart_rxcnt++;
	}
}

